"""
Access legacy Aleph servicelayer archive read-only without `servicelayer`
dendency
"""

from enum import StrEnum
from functools import cache

from anystore.store import BaseStore, get_store
from pydantic_settings import BaseSettings, SettingsConfigDict

from openaleph_procrastinate.exceptions import ArchiveFileNotFound


class ArchiveType(StrEnum):
    file = "file"
    s3 = "s3"
    gcs = "gcs"


class ArchiveSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="archive_")

    type: ArchiveType = ArchiveType.file
    bucket: str | None = None
    path: str | None = None
    endpoint_url: str | None = None


@cache
def get_archive() -> BaseStore:
    settings = ArchiveSettings()
    if settings.type == ArchiveType.s3:
        return get_store(f"s3://{settings.bucket}")
    if settings.type == ArchiveType.gcs:
        return get_store(f"gcs://{settings.bucket}")
    return get_store(uri=settings.path)


def lookup_key(content_hash: str) -> str:
    archive = get_archive()
    for key in archive.iterate_keys(prefix=content_hash):
        return key
    raise ArchiveFileNotFound(f"Key does not exist: `{content_hash}`")



"""
Helper functions to access Archive and FollowTheMoney data within Jobs
"""

from contextlib import contextmanager
from pathlib import Path
from typing import BinaryIO, Generator

from anystore.store.virtual import get_virtual
from followthemoney.proxy import EntityProxy
from ftmstore import get_dataset
from ftmstore.loader import BulkLoader

from openaleph_procrastinate.exceptions import EntityNotFound
from openaleph_procrastinate.legacy.archive import get_archive, lookup_key
from openaleph_procrastinate.settings import settings

OPAL_ORIGIN = "openaleph_procrastinate"


@contextmanager
def get_localpath(dataset: str, content_hash: str) -> Generator[Path, None, None]:
    """
    Load a file from the archive and store it in a local temporary path for
    further processing. The file is cleaned up after leaving the context.

    !!! danger
        This is not tested.
    """
    archive = get_archive()
    key = lookup_key(content_hash)
    store = get_virtual()
    path = store.download(key, archive)
    try:
        yield Path(path)
    finally:
        store.cleanup(path)


@contextmanager
def open_file(dataset: str, content_hash: str) -> Generator[BinaryIO, None, None]:
    """
    Load a file from the archive and store it in a local temporary path for
    further processing. Returns an open file handler. The file is closed and
    cleaned up after leaving the context.

    !!! danger
        This is not tested.
    """
    archive = get_archive()
    key = lookup_key(content_hash)
    with archive.open(key) as handler:
        try:
            yield handler
        finally:
            handler.close()


def load_entity(dataset: str, entity_id: str) -> EntityProxy:
    """
    Retrieve a single entity from the store.
    """
    store = get_dataset(dataset, database_uri=settings.ftm_store_uri)
    entity = store.get(entity_id)
    if entity is None:
        raise EntityNotFound(f"Entity `{entity_id}` not found in dataset `{dataset}`")
    return entity


@contextmanager
def entity_writer(dataset: str) -> Generator[BulkLoader, None, None]:
    """
    Get the `ftmstore.dataset.BulkLoader` for the given `dataset`. The entities
    are flushed when leaving the context.
    """
    store = get_dataset(
        dataset, origin=OPAL_ORIGIN, database_uri=settings.ftm_store_uri
    )
    loader = store.bulk()
    try:
        yield loader
    finally:
        loader.flush()



from openaleph_procrastinate import logging
from openaleph_procrastinate.logging import get_logger
from openaleph_procrastinate.settings import settings

log = get_logger(__name__)


class InvalidJob(Exception):
    pass


class ArchiveFileNotFound(Exception):
    pass


class EntityNotFound(Exception):
    pass


class ErrorHandler:
    def __init__(self, logger: logging.BoundLogger | None = None) -> None:
        self.log = logger or log

    def __enter__(self):
        pass

    def __exit__(self, e, msg, _):
        if e is not None:
            if settings.debug:
                raise e
            self.log.error(f"{e.__name__}: {msg}")



from typing import Annotated, Optional

import typer
from anystore.cli import ErrorHandler
from anystore.io import smart_stream_json
from ftmq.io import smart_stream_proxies
from rich import print

from openaleph_procrastinate import __version__, model, tasks
from openaleph_procrastinate.app import app
from openaleph_procrastinate.logging import configure_logging
from openaleph_procrastinate.settings import settings

cli = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=settings.debug)

DEFAULT_QUEUE = "default"
DEFAULT_TASK = "openaleph_procrastinate.tasks.dummy_task"

OPT_INPUT_URI = typer.Option("-", "-i", help="Input uri, default stdin")
OPT_DATASET = typer.Option(..., "-d", help="Dataset")
OPT_QUEUE = typer.Option(DEFAULT_QUEUE, "-q", help="Queue name")
OPT_TASK = typer.Option(
    "openaleph_procrastinate.tasks.dummy_task", "-t", help="Task module path"
)


@cli.callback(invoke_without_command=True)
def cli_opal_procrastinate(
    version: Annotated[Optional[bool], typer.Option(..., help="Show version")] = False,
):
    if version:
        print(__version__)
        raise typer.Exit()
    configure_logging()


@cli.command()
def defer_entities(
    input_uri: str = OPT_INPUT_URI,
    dataset: str = OPT_DATASET,
    queue: str = OPT_QUEUE,
    task: str = OPT_TASK,
):
    """
    Defer jobs for a stream of proxies
    """
    with ErrorHandler(), app.open():
        for proxy in smart_stream_proxies(input_uri):
            job = model.DatasetJob.from_entity(
                dataset=dataset, queue=queue, task=task, entity=proxy
            )
            job.defer(app)


@cli.command()
def defer_jobs(input_uri: str = OPT_INPUT_URI):
    """
    Defer jobs from an input json stream
    """
    with ErrorHandler(), app.open():
        for data in smart_stream_json(input_uri):
            job = tasks.unpack_job(data)
            job.defer(app)



from pathlib import Path
from typing import Any, BinaryIO, ContextManager, Generator, Iterable, Self, TypeAlias

from followthemoney import model
from followthemoney.proxy import EntityProxy
from ftmstore.loader import BulkLoader
from procrastinate.app import App
from pydantic import BaseModel, ConfigDict, Field

from openaleph_procrastinate import helpers


class EntityFileReference(BaseModel):
    """
    A file reference (via `content_hash`) to a servicelayer file from an entity
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    dataset: str
    content_hash: str
    entity: EntityProxy

    def open(self: Self) -> ContextManager[BinaryIO]:
        """
        Open the file attached to this job

        !!! danger
            This is not tested.
        """
        return helpers.open_file(self.dataset, self.content_hash)

    def get_localpath(self: Self) -> ContextManager[Path]:
        """
        Get a temporary path for the file attached to this job

        !!! danger
            This is not tested.
        """
        return helpers.get_localpath(self.dataset, self.content_hash)


class Stage(BaseModel):
    """Define an arbitrary (next) stage for a job"""

    # model: str
    queue: str
    task: str

    def make_job(self: Self, job: "AnyJob") -> "AnyJob":
        """Create a new job for this stage from a previous one"""
        job.queue = self.queue
        job.task = self.task
        return job


class Job(Stage):
    """
    A job with arbitrary payload
    """

    payload: dict[str, Any]
    stages: list[Stage] = Field(default=[])

    @property
    def context(self) -> dict[str, Any]:
        """Get the context from the payload if any"""
        return self.payload.get("context") or {}

    def defer(self: Self, app: App) -> None:
        """Defer this job"""
        data = self.model_dump(mode="json")
        app.configure_task(name=self.task, queue=self.queue).defer(**data)


class DatasetJob(Job):
    """
    A job with arbitrary payload bound to a `dataset`.
    The payload always contains an iterable of serialized `EntityProxy` objects
    in the `entities` key. It may contain other payload context data in the
    `context` key.

    There are helpers for accessing archive files or loading entities.
    """

    dataset: str

    def get_writer(self: Self) -> ContextManager[BulkLoader]:
        """Get the writer for the dataset of the current entity"""
        return helpers.entity_writer(self.dataset)

    def get_entities(self) -> Generator[EntityProxy, None, None]:
        """
        Get the entities from the payload
        """
        assert "entities" in self.payload, "No entities in payload"
        for data in self.payload["entities"]:
            yield model.get_proxy(data)

    def load_entities(self: Self) -> Generator[EntityProxy, None, None]:
        """Load the entities from the store to refresh it to the latest data"""
        assert "entities" in self.payload, "No entities in payload"
        for data in self.payload["entities"]:
            yield helpers.load_entity(self.dataset, data["id"])

    # Helpers for file jobs that access the servicelayer archive

    def get_file_references(self) -> Generator[EntityFileReference, None, None]:
        """Get file references per entity from this job"""
        for entity in self.get_entities():
            for content_hash in entity.get("contentHash"):
                yield EntityFileReference(
                    dataset=self.dataset, entity=entity, content_hash=content_hash
                )

    # Helpers for creating entity jobs

    @classmethod
    def from_entities(
        cls, dataset: str, queue: str, task: str, entities: Iterable[EntityProxy]
    ) -> Self:
        """Make a job to process entities for a dataset"""
        return cls(
            dataset=dataset,
            queue=queue,
            task=task,
            payload={"entities": [e.to_full_dict() for e in entities]},
        )

    @classmethod
    def from_entity(
        cls, dataset: str, queue: str, task: str, entity: EntityProxy
    ) -> Self:
        """Make a job to process an entity for a dataset"""
        return cls.from_entities(
            dataset=dataset, queue=queue, task=task, entities=[entity]
        )


AnyJob: TypeAlias = Job | DatasetJob



import procrastinate

from openaleph_procrastinate.settings import settings


def make_app(tasks_module: str) -> procrastinate.App:
    return procrastinate.App(
        connector=procrastinate.PsycopgConnector(
            conninfo=settings.db_uri,
        ),
        import_paths=[tasks_module],
    )


app = make_app("openaleph_procrastinate.tasks")



import logging
import os
import sys
from logging import Filter, LogRecord
from typing import Any, Dict, List

import structlog
from banal import as_bool
from structlog.contextvars import merge_contextvars
from structlog.dev import ConsoleRenderer, set_exc_info
from structlog.processors import (  # format_exc_info,
    JSONRenderer,
    TimeStamper,
    UnicodeDecoder,
    add_log_level,
)
from structlog.stdlib import (
    BoundLogger,
    LoggerFactory,
    ProcessorFormatter,
    add_logger_name,
)
from structlog.stdlib import get_logger as get_raw_logger

LOG_JSON = as_bool(os.environ.get("LOG_JSON"))
LOG_LEVEL = os.environ.get("LOG_LEVEL", "info").upper()


def get_logger(name: str, *args, **kwargs) -> BoundLogger:
    return get_raw_logger(name, *args, **kwargs)


def configure_logging(level: int = logging.INFO) -> None:
    """Configure log levels and structured logging"""
    shared_processors: List[Any] = [
        add_log_level,
        add_logger_name,
        # structlog.stdlib.PositionalArgumentsFormatter(),
        # structlog.processors.StackInfoRenderer(),
        merge_contextvars,
        set_exc_info,
        TimeStamper(fmt="iso"),
        # format_exc_info,
        UnicodeDecoder(),
    ]

    if LOG_JSON:
        # shared_processors.append(format_exc_info)
        shared_processors.append(format_json)
        formatter = ProcessorFormatter(
            foreign_pre_chain=shared_processors,
            processor=JSONRenderer(),
        )
    else:
        formatter = ProcessorFormatter(
            foreign_pre_chain=shared_processors,
            processor=ConsoleRenderer(
                exception_formatter=structlog.dev.plain_traceback
            ),
        )

    processors = shared_processors + [
        ProcessorFormatter.wrap_for_formatter,
    ]

    # configuration for structlog based loggers
    structlog.configure(
        cache_logger_on_first_use=True,
        # wrapper_class=AsyncBoundLogger,
        wrapper_class=BoundLogger,
        processors=processors,
        context_class=dict,
        logger_factory=LoggerFactory(),
    )

    # handler for low level logs that should be sent to STDERR
    out_handler = logging.StreamHandler(sys.stderr)
    out_handler.setLevel(level)
    out_handler.addFilter(_MaxLevelFilter(logging.WARNING))
    out_handler.setFormatter(formatter)
    # handler for high level logs that should be sent to STDERR
    error_handler = logging.StreamHandler(sys.stderr)
    error_handler.setLevel(logging.ERROR)
    error_handler.setFormatter(formatter)

    root_logger = logging.getLogger()
    root_logger.setLevel(LOG_LEVEL)
    root_logger.addHandler(out_handler)
    root_logger.addHandler(error_handler)


def format_json(_: Any, __: Any, ed: Dict[str, str]) -> Dict[str, str]:
    """Stackdriver uses `message` and `severity` keys to display logs"""
    ed["message"] = ed.pop("event")
    ed["severity"] = ed.pop("level", "info").upper()
    return ed


class _MaxLevelFilter(Filter):
    def __init__(self, highest_log_level: int) -> None:
        self._highest_log_level = highest_log_level

    def filter(self, record: LogRecord) -> bool:
        return record.levelno <= self._highest_log_level
